'use strict';
import util from 'util';
import { parse, resolve, normalize } from 'path';
import { createHash } from 'crypto';
/*
减少语句路径
少用pop,push 改为i++--操作和[length]=new
频繁new的对象改为inherit2方法 尽管inherit2方法无法私有对象
多用三元？少用if
多用空方法 少用if
除非必须 少用delete
使用 || 代替 getValue 或者3元
使用for if(i++>1) 判断是否为空
使用substr lastIndexOf 代替split
尽量使用原生方法（native）代替自己的方法
*/
export const environment = {
    os: process.env.COMMAND_MODE || process.env.OS,
    splitChar: (process.env.COMMAND_MODE || process.env.OS).toLowerCase().indexOf('win') >= 0 ? "\\" : "/"
};
/**
 * 采用代理模式给类和类内方法添加属性
 * @param(调用时的原方法对象,键值,键值上的一个新值)
 * @returns decorator
 */
export const prop = (target, k, v) => {
    target[k] = target[k] || [];
    target[k].push(v);
};
/**
 * 添加link属性用于连接方法需求
 * @param (link)
 */
export const link = url => { return (target) => prop(target, 'link', url); };
/**
 * 添加多个desc属性用于动态显示方法描述
 * @param (desc)
 */
export const desc = des => { return (target) => prop(target, 'desc', des); };
/**
 * 判断对象是否有效 含有 not in (null,undefined,'null','',\s+,0,false)
 */
export const isValid = data => (!!data && (data.replace == undefined || (data != 'null' && data.replace(/\s/g, '').length > 0)));
/**
 * 三元表达式，简化 多数情况可以用默认值或者 value || defaultValue 替换，但是当其为null时则不好判断
 */
export const getValue = (data, defaultData) => isValid(data) ? data : defaultData;
/**
 * 私有方法，返回_,__命名的一个自身对象和一个私有对象。一般在类方法第一句使用
 * @param(this,__)
 * @return({this,private})
 */
export const pris = () => { const key = Symbol('private'); return (obj, priv) => { obj[key] = (priv ? merge(obj[key], priv) : obj[key]) || {}; return { '_': obj, '__': obj[key] }; }; };

/**
 * 字符串格式化，使用<%k.v%>,{k.v}等等方式进行递进的属性替换，当字符串模板不可用时使用。
 */
export const format = function(s, o) {
    if (!s || !o) { return getValue(s, ''); }
    if (s.indexOf('<%=') >= 0) return __.funrep(s, o, /<%=[^(%>)]+%>/gi, /<%=/g, /%>/g);
    if (s.indexOf('{') >= 0) return __.funrep(s, o, /\{[^(})]+\}/gi, /\{/g, /\}/g);
    return s;
};
/**
 * 返回私有StringBuilder对象实例，含有
 */
export const sb = () => new __.StringBuilder();
export const showException = (name = '', e) => {
    if (isValid(e)) {
        name += ("\r\nmessage:" + e.message + (e.stack ? ("\r\nstack:" + e.stack + (e.fileName ? ("\r\nfile:" + e.fileName) : '') + (e.lineNumber ? ("\r\nlineNumber:" + e.lineNumber) : '')) : (e.description ? ("\r\ndescription:" + e.description) : "")));
        console.error(name);
        console.log(name);
    }
};
export const showEx = showException;
export const tryF = (func, errcall) => new Proxy(func, __.tryHandle(errcall));
export const tryC = (func, errcall) => tryF(func, errcall)();
export const tryC2 = (err, func, errcall) => err ? (errcall || __.showEx)(e) : tryC(func, errcall);
export const watch = (restart) => {
    if (!__.start || restart) {
        __.start = new Date();
        console.log('VESH.watch开始' + __.start);
    } else {
        console.log('VESH.watch 持续了:' + __.start.diff('ms', new Date()));
    }
}
export const isArray = util.isArray;
export const once = (func, timeout) => {
    timeout = timeout || 1;
    if (func.timeoutID) {
        clearTimeout(func.timeoutID);
    }
    if (timeout == 1)
        process.nextTick(() => tryC(func));
    else
        func.timeoutID = setTimeout(() => tryC(func), timeout);
};
//exp的非链式操作与链式操作，非链式不保证执行顺序，链式保证执行顺序，对象的随机并发操作与顺序执行，数组的map随机并发操作与顺序执行
//使用once满足非promise截止的请求，循环全部是promise截止的r
//根据执行的源头不同分为 whileC(链式、非链式，promise),forC(链式、非链式，promise),each(链式、非链式，promise),next(链式 promise),finalC(非链式 promise) 不同种类 根据链式与非链式，
//保证异步的同步执行
//保证当直接返回true或者call(null,true)时中断循环
export const whileC = (exp, func, isSequ = false) => isSequ ? __.sequfunc(exp, func, exp()) : __.disorderfunc(exp, func, exp());
export const each = (data, func, isSequ = false) => isSequ ? __.sequEach(data, func) : __.disorderEach(data, func);
export const forC = (obj, func, isSequ = false) => isSequ ? __.sequFor(obj, func) : __.disorderFor(obj, func);
export const finalC = (...funs) => {
    const data = {},
        finalF = funs.pop();
    return each(funs, v => callback2(v, null, data)).then(() => callback2(finalF, null, data));
};
export const next = (...funs) => {
    const data = {};
    return each(funs, v => callback2(v, null, data), true);
};
/**
 * 异步改同步请求
 * 兼容直接返回值，返回一个Promise或者返回undefined 最后一个参数调用call(e,data) 三种情况
 * @param {*} args 
 */
export const callback2 = function(...args) {
    return new Promise(async(re, rj) => {
        try {
            const call = (e, data) => e ? rj(e) : re(data);
            const func = args[0];
            const _this = args[1] || null;
            const paras = args.slice(2).concat([call]);
            let data = Reflect.apply(func, _this, paras); //this 不对 Reflect.apply(func, null, paras);
            if (data && data.then) data = await data || false;
            switch (typeof(data)) {
                case 'function':
                    data(call);
                case 'undefined':
                    break;
                default:
                    call(null, data);
                    break;
            }
        } catch (e) {
            rj(e);
        }
    });
};
/**
 * 仅适用最后一个参数为call的情况
 * @param {*} args 
 */
export const callback = function(...args) {
    const func = args[0];
    args[0] = function(...args2) {
        const that = this;
        Reflect.apply(func, that, args2);
    };
    return callback2(...args);
};
export const random = () => parseInt('' + (new Date()).getTime() + (__.index++));
/**
 * 获取字符串的哈希值
 * @param {String} str
 * @param {Boolean} caseSensitive
 * @return {Number} hashCode
 */
export const hash = (str, caseSensitive = false) => {
    str = str.toString();
    if (!caseSensitive) {
        str = str.toLowerCase();
    }
    // 1315423911=b'1001110011001111100011010100111'
    let hash = 1315423911,
        i, ch;
    for (i = str.length - 1; i >= 0; i--) {
        ch = str.charCodeAt(i);
        hash ^= ((hash << 5) + ch + (hash >> 2));
    }
    return (hash & 0x7FFFFFFF);
};
/**
 * 获取字符串的sha1 base64 哈希值
 * @param {String} str
 * @param {Boolean} caseSensitive
 * @return {Number} hashCode
 */
export const hash2 = (str) => createHash('sha256').update(str).digest('base64');
export const formatPrice = function(number, decimals, dec_point, thousands_sep) {
    number = (number + '').replace(/[^0-9+-Ee.]/g, '');
    let n = !isFinite(+number) ? 0 : +number,
        prec = !isFinite(+decimals) ? 2 : Math.abs(decimals),
        sep = (typeof thousands_sep === 'undefined') ? ',' : thousands_sep,
        dec = (typeof dec_point === 'undefined') ? '.' : dec_point,
        s = '',
        toFixedFix = function(n, prec) {
            const k = Math.pow(10, prec);
            return '' + Math.round(n * k) / k;
        };
    // Fix for IE parseFloat(0.55).toFixed(0) = 0;
    s = (prec ? toFixedFix(n, prec) : '' + Math.round(n)).split('.');
    if (s[0].length > 3) {
        s[0] = s[0].replace(/B(?=(?:fd{3})+(?!d))/g, sep);
    }
    if ((s[1] || '').length < prec) {
        s[1] = s[1] || '';
        s[1] += new Array(prec - s[1].length + 1).join('0');
    }
    return s.join(dec);
};
/**
 * 只更新标点符号不更新全半角文字的encode
 * @param {String} html
 * @return {String} encodeURIComponent之后的字段
 */
export const encHtml = (html = '') => html.replace(/[\r\n]+/g, '>v>j>').replace(/\s+/g, ' ').replace(/>v>j>/g, '\r\n').replace(__.encreg, a => encodeURIComponent(a));
export const decHtml = (html = '') => decodeURIComponent(html);
export const toJsonString = JSON.stringify;
export const json = JSON.parse;
export const toTJson = (data = []) => {
    if (!isArray(data)) data = [data];
    const res = [];
    if (data[0] && isArray(data[0])) {
        for (const i in data) {
            res[i] = toTJson(data[i]);
        }
    } else {
        //找到第一个不是数组的实例对象进行处理
        res[0] = [];
        for (const k in data[0]) {
            res[0][res[0].length] = "\"" + k + "\"";
        }
        for (const v in data) {
            const obj = [];
            for (const k2 in data[v])
                obj[obj.length] = __.getTJsonValue(data[v][k2]);
            res[res.length] = (obj);
        }
    }
    return res;
};
export const getType = (x = null) => {
    if (x == null) {
        return "null";
    }
    const t = typeof x;
    if (t != "object" && t != 'Object') {
        return t;
    }
    if (isArray(x)) {
        return 'Array';
    }
    let c = Object.prototype.toString.apply(x);
    c = c.substring(8, c.length - 1);
    if (c != "Object") {
        return c;
    }
    if (x.constructor == Object) {
        return c;
    }
    if (x.prototype && "classname" in x.prototype.constructor &&
        typeof x.prototype.constructor.classname == "string") {
        return x.constructor.prototype.classname;
    }
    return "ukObject";
};
export const merge = (...argu) => {
    if (argu.length < 2) { return argu[0] ? argu[0] : {} };
    if (argu.length > 0 && true == argu[argu.length - 1]) {
        let _ = argu[0];
        for (let i2 = 1; i2 < argu.length; i2++)
            _ = __.merge(_, argu[i2]);
        return _;
    } else {
        let _ = {};
        for (let i2 = 0; i2 < argu.length; i2++)
            _ = __.merge(_, argu[i2]);
        return _;
    }
}
export const inherits = function(parent, args) {
    //绕过了parent的构造函数，重新链接原型链条
    const _temp = (function() {
        const F = function() {};
        util.inherits(F, parent);
        //F.prototype = parent.prototype;
        F.prototype.isF = true;
        return new F();
    })();
    _temp.constructor = parent;
    if (!this.prototype) {
        //这里确认是实例
        //确定是打断了原型链 使得this的原型为Object		
        parent.apply(this, args);
        //从新接驳原型链 使得原型链上的prototype都设置到最早的类的prototype上了
        if (this.__proto__ && !this.__proto__.isF) {
            util.inherits(this.__proto__.constructor, _temp.__proto__.constructor);
            //this.__proto__.constructor.prototype = _temp.__proto__.constructor.prototype;
        }
        //son.prototype = _temp; //这里可以分层 但是会使得prototype实例变了又变 废弃
        this.__proto__ = _temp;
        //父类方法只能找到静态方法
        this.base = this.__proto__.constructor.prototype;
    } else {
        console.log('如果失败，需要配合子类构造函数中使用parent.apply(this,[***])');
        //这里确认是类定义
        this.prototype = _temp;
    }
};
export const create2 = function(type, args) {
    if (typeof(type) == 'function') {
        return new type(...args);
    } else showException('请传入类定义');
};
export const create = function(type, args) {
    if (typeof(type) == 'function') {
        args = Array.isArray(args) ? args : [args];
        let ret = '(new type(';
        if (Array.isArray(args)) {
            for (const i in args) {
                ret += 'args[' + i + '],'
            }
            if (args.length > 0) {
                ret = ret.substr(0, ret.length - 1);
            }
        }
        return eval(ret + '))');
    } else showException('请传入类定义');
};
/**
 * 分别尝试全路径.js/全路径.njs/全路径加载类 并对处理结果进行缓存 会导致热更新失败错误!
 * @param {文件上级路径} path 
 * @param {类名} base 
 */
export const include = function(path, base = null) {
    const paths = `${path}|${base}`;
    return __.include[paths] || (function() {
        if (__.unclude[paths]) throw new Error(`加载${paths}失败!`);
        __.include[paths] = (function() {
            try {
                if (path.startWith(environment.splitChar)) {
                    base = (base ? format(base) : parse(resolve(normalize(process.mainModule.filename))).dir);
                    if (environment.os & 'windows') {
                        //windows
                        base = base.replace(/\//g, '\\');
                        path = path.trim('/').trim('\\').replace(/\//g, '\\');
                    } else {
                        //linux
                        base = base.replace(/\\\\/g, '/');
                        path = path.trim('\\').trim('/').replace(/\\\\/g, '/');
                    }
                    try {
                        try {
                            return require(`${base}${environment.splitChar}${path}.js`);
                        } catch (e2) {
                            return require(`${base}${environment.splitChar}${path}.njs`);
                        }
                    } catch (e3) {
                        console.error(e3.stack);
                        return require(path);
                    }
                } else return require(path);
            } catch (e) {
                console.error(e.stack);
                __.unclude[paths] = true;
                throw new Error(`加载${paths}失败!${e.stack}`);
            }
        })();
        return __.include[paths];
    })();
};
export const applyCAE = function(S) {
    //功能出现严重问题 无法在ES6下使用
    S[__.settings] = {};
    S[__.exsettings] = {};
    //获取不存在就配置
    S.getSettings = function(key, data) {
        if (!isValid(S[__.settings][key])) {
            if (isValid(S[__.exsettings][key])) {
                S[__.settings][key] = merge(getValue(data, {}), S[__.exsettings][key]);
                delete S[__.exsettings][key];
            } else
                S[__.settings][key] = getValue(data, {});
        }
        return S[__.settings][key];
    };
    //扩展默认配置
    S.extendSettings = function(key, data) {
        if (isValid(S[__.settings][key])) {
            S[__.settings][key] = merge(S[__.settings][key], data);
        } else {
            if (S.exSettings[key]) {
                S[__.exsettings][key] = merge(S[__.exsettings][key], getValue(data, {}));
            } else {
                S[__.exsettings][key] = getValue(data, {});
            }
        }
    };
    S.clearSettings = function() { S[__.settings] = {}; };
    S.registCommand = function(name, func) {
        const comms = S.getSettings('comms', []);
        const data = comms[name];
        if (isValid(data) && typeof(data) != 'function') {
            func.apply(null, data);
        }
        comms[name] = func;
    };
    /*
    V用于调用被调用页面注册的命令以处理异步命令调用，当命令尚未注册而已经被调用时，参数会先被缓存下来，然后当命令注册时，已知的参数再被调用。
    --案例
    S.callCommand('showXXList',[{id:1}])
    */
    S.callCommand = function(name, data) {
        const caller = null;
        const comms = S.getSettings('comms', []);
        const func = comms[name];
        data = isArray(data) ? data : [data];
        if (isValid(func) && typeof(func) == 'function') {
            once(function() { func.apply(caller, data); });
        } else {
            comms[name] = data;
        }
    };
    /*
    用来判断是否调用页面,当已经调用过(part)，返回true,否则返回false;
    --案例
    if (!S.hasCommand('editor.open')) S.part("/FileServer/layout/editor/editor.htm");
    */
    S.hasCommand = function(name) {
        const comms = S.getSettings('comms', []);
        const func = comms[name];
        return (isValid(func) && typeof(func) == 'function');
    };

    /*
    仅限iframe方式调用时，先取消原页面添加的方法
    //业务逻辑深度交叉，iframe落后的控件连接方式时使用
    一定要在part前
    --案例
    S.cleanCommand('editor.open');
    S.part("/FileServer/layout/editor/editor.htm",null,"iframe",function(){});
    */
    S.cleanCommand = function(name) {
        const comms = S.getSettings('comms', []);
        delete comms[name];
    };
    /*
    V用于被调用页面注册命令以处理异步命令调用,当命令尚未注册而已经被调用时，参数会先被缓存下来，然后当命令注册时，已知的参数再被调用。
    并约定1分钟内 允许注册者多次被触发
    --案例
    S.registEvent('showXXList',getData),S.registEvent(['showXXList',''],getData)
    */
    S.registEvent = function(name, func, isTop) {
        const fun = function(name, func, isTop) {
            const events = S.getSettings('events', []);
            let funs = events[name];
            if (!isValid(funs)) {
                funs = [];
                events[name] = funs;
            }
            if (typeof(func) == 'function') {
                if (isTop && !funs.top) {
                    funs.top = func;
                    funs.unshift(func);
                } else {
                    if (isTop && funs.top) { showException('S.registEvent:' + name + ' 事件已经有订阅者被置顶!'); }
                    funs.push(func);
                }
                let ecall = S.getSettings('eventcall', {});
                ecall = ecall[name] ? ecall[name] : {};
                if (ecall.time && ecall.time >= (new Date().getTime())) {
                    once(function() {
                        func.apply(ecall.caller, ecall.data);
                    });
                }
            }
        };
        if (isArray(name)) {
            each(name, function(v) {
                fun(v, func, isTop);
                return false;
            }, true);
        } else {
            fun(name, func, isTop);
        }
    };
    /*
    V用于调用被调用页面注册的事件以处理异步命令调用，当命令尚未注册而已经被调用时，参数会先被缓存下来，然后当命令注册时，已知的参数再被调用。
    并约定1分钟内 允许注册者多次被触发
    --案例
    S.callEvent('showXXList',[{id:1}])
    */
    S.callEvent = function(name, data) {
        const caller = null;
        const events = S.getSettings('events', []);
        const funs = events[name];
        data = isArray(data) ? data : [data];
        if (isValid(funs) && isArray(funs)) {
            each(funs, function(func) {
                //报错不下火线
                tryC(function() {
                    func.apply(caller, data);
                });
            });
        }
        let ecall = S.getSettings('eventcall', {});
        if (!ecall[name]) { ecall[name] = {}; }
        ecall = ecall[name];
        ecall.time = new Date().add('n', 1).getTime();
        ecall.data = data;
        ecall.caller = caller;
    };
    /*
    用来判断是否调用页面,当已经调用过(part)，返回true,否则返回false;
    --案例
    if (!S.hasEvent('editor.open')) S.part("/FileServer/layout/editor/editor.htm");
    */
    S.hasEvent = function(name) {
        const events = S.getSettings('events', []);
        const funs = events[name];
        if (isValid(funs) && isArray(funs)) {
            return true;
        }
        return false;
    };

    /*
    仅限iframe方式调用时，先取消原页面添加的方法
    //业务逻辑深度交叉，iframe落后的控件连接方式时使用
    一定要在part前
    --案例
    S.cleanEvent('editor.open');
    S.part("/FileServer/layout/editor/editor.htm",null,"iframe",function(){});
    */
    S.cleanEvent = function(name) {
        const events = S.getSettings('events', []);
        delete events[name];
    };
};
export default { environment, prop, link, desc, isValid, getValue, format, sb, showEx, showException, include, random, tryC, tryC2, watch, isArray, once, whileC, forC, each, finalC, next, callback, callback2, encHtml, decHtml, toJsonString, json, merge, inherits, create, create2, applyCAE, pris, hash, hash2, toTJson };

//const symbol = Symbol;
const pri = pris();
const __ = {
    settings: Symbol('settings'),
    exsettings: Symbol('exsettings'),
    encreg: new RegExp('<|>|~|(\r\n)|!|@|#|\\$|%|\\^|;|\\*|\\(|\\)|_|\\+|\\{|\\}|\\||:|\"|\\?|`|\\-|=|\\[|\\]|\\\|;|\'|,|\\.|/|，|；', 'g'),
    funrep: function(s, o, reg, lreg, rreg) {
        return s.replace(reg, function(word) {
            const key = '' + word.replace(lreg, '').replace(rreg, '');
            return o[key] || '';
        });
    },
    /**
     * StringBuilder nodejs版本 用于连续操作字符串对象而不经常移动数组 比较准确快捷的修改字符串内容
     */
    StringBuilder: class {
        constructor() {
            const that = this;
            const { _, __ } = pri(that, {
                data: [],
                length: 0,
                append: (str = "") => {
                    const { _, __ } = pri(that);
                    __.data[__.data.length] = str;
                    __.length += str.length;
                }
            });
        }
        get length() {
            return pri(this).__.length;
        }
        append(str) {
            const { _, __ } = pri(this);
            str = isValid(str) ? __.append(str) : '';
            return this;
        }
        appendFormat(fmt, data) {
            return this.append(format(fmt, data));
        }
        insert(start, data) {
            const { _, __ } = pri(this);
            const str = _.toString();
            __.data = [str.substr(0, start), data, str.substr(start)];
            __.length = str.length + data.length;
            return this;
        }
        insertFormat(start, fmt, data) {
            this.insert(start, format(fmt, data));
        }
        remove(start, length) {
            const { _, __ } = pri(this);
            let str = _.toString();
            __.data = [str.substr(0, start), str.substr(start + length)];
            __.length = Math.max(0, str.length - length);
            return _;
        }
        toString() {
            const { _, __ } = pri(this);
            __.data = [__.data.join('')];
            return __.data[0];
        }
        clear() {
            let ret = this.toString();
            const { _, __ } = pri(this);
            __.data = [];
            __.length = 0;
            return ret;
        }
    },
    showEx: (e) => showEx('', e),
    tryHandle: function(errcall = __.showEx) {
        return {
            apply: function(target, context, proxy) {
                try {
                    return Reflect.apply(...arguments);
                } catch (e) {
                    try { errcall(e); } catch (e2) { __.showEx(e); }
                }
            }
        };
    },
    start: null,
    emptyfunc: () => true,
    sequfunc: async(exp, func, val) => {
        let isStop = false;
        while (!isStop && val != undefined) {
            isStop = true == await callback2(func, null, val);
            val = exp();
        }
        return true;
    },
    disorderfunc: (exp, func, val) => {
        let data = [];
        while (val != undefined) {
            data[data.length] = callback2(func, null, val);
            val = exp();
        }
        return Promise.all(data);
    },
    sequEach: (data = [], func) => {
        let i = 0;
        return __.sequfunc(() => data[i++], func, data[i++]);
    },
    disorderEach: (data = [], func) => Promise.all(data.map(k => callback2(func, null, k))),
    sequFor: async(obj = {}, func) => {
        for (const k in obj) {
            if (true == await callback2(func, obj, k, obj[k]))
                return true;
        }
        return true;
    },
    disorderFor: (obj = {}, func) => {
        let data = [];
        for (const k in obj) data[data.length] = callback2(func, obj, k, obj[k]);
        return Promise.all(data);
    },
    index: 0,
    getTJsonValue: p => {
        switch (typeof(p)) {
            case 'number':
            case 'boolean':
                return p;
            case 'undefined':
            case 'object':
                return "\"\"";
            default:
                return "\"" + p + "\"";
        }
    },
    clone: source => {
        switch (getType(source)) {
            case 'Object':
            case 'object':
                return __.merge({}, source);
            case 'array':
            case 'Array':
                const aim = [];
                for (const i in source) {
                    aim.push(__.clone(source[i]));
                }
                return aim;
            default:
                return source;
        }
    },
    merge: (aim, source) => {
        if (!(typeof(source) == 'object' && !isArray(source))) { return aim; }
        for (const i in source) {
            if (source[i] != undefined) {
                if (!isValid(aim[i])) {
                    aim[i] = __.clone(source[i]);
                } else {
                    switch (getType(aim[i])) {
                        case 'object':
                        case 'Object':
                            __.merge(aim[i], source[i]);
                            break;
                        case 'Array':
                            //处理数组
                            let hasmergeIndex = false;
                            for (let i3 = 0, k = source[i][i3]; i3 < source[i].length; i3++, k = source[i][i3]) {
                                if (typeof(k.mergeIndex) == "number") {
                                    hasmergeIndex = true;
                                    if (aim[i].length < (k.mergeIndex + 1)) {
                                        aim[i].push(k);
                                    } else {
                                        aim[i][i3] = __.merge(aim[i][i3], k);
                                    }
                                } else if (typeof(k.moveIndex) == "number") {
                                    hasmergeIndex = true;
                                    aim[i].splice(k.moveIndex, 0, k);
                                }
                            }
                            if (!hasmergeIndex) {
                                aim[i] = __.clone(source[i]);
                            }
                            break;
                        default:
                            aim[i] = source[i];
                            break;
                    }
                }
            }
        }
        return aim;
    },
    include: {},
    unclude: {}
}; {
    //日期函数处理
    Date.prototype.add = function(interval, number) {
            const d = new Date(this.getTime());
            const k = { 'y': 'FullYear', 'q': 'Month', 'm': 'Month', 'w': 'Date', 'd': 'Date', 'h': 'Hours', 'n': 'Minutes', 's': 'Seconds', 'ms': 'MilliSeconds' };
            const n = { 'q': 3, 'w': 7 };
            eval('d.set' + k[interval] + '(d.get' + k[interval] + '()+' + ((n[interval] || 1) * number) + ')');
            return d;
        }
        /* 计算两日期相差的日期年月日等 new Date().diff('h',new Date().add('d',1)); */
    Date.prototype.diff = function(interval, objDate2) {
            const d = this,
                i = {},
                t = d.getTime(),
                t2 = objDate2.getTime();
            i['y'] = objDate2.getFullYear() - d.getFullYear();
            i['q'] = i['y'] * 4 + Math.floor(objDate2.getMonth() / 4) - Math.floor(d.getMonth() / 4);
            i['m'] = i['y'] * 12 + objDate2.getMonth() - d.getMonth();
            i['ms'] = objDate2.getTime() - d.getTime();
            i['w'] = Math.floor((t2 + 345600000) / (604800000)) - Math.floor((t + 345600000) / (604800000));
            i['d'] = Math.floor(t2 / 86400000) - Math.floor(t / 86400000);
            i['h'] = Math.floor(t2 / 3600000) - Math.floor(t / 3600000);
            i['n'] = Math.floor(t2 / 60000) - Math.floor(t / 60000);
            i['s'] = Math.floor(t2 / 1000) - Math.floor(t / 1000);
            return i[interval];
        }
        /* 计算两日期相差的日期年月日等 new Date().diff('h',new Date().add('d',1)); */
    Date.prototype.sub = function(interval, objDate2) {
            return -1 * Date.prototype.diff.apply(this, [interval, objDate2]);
        }
        /* 计算两日期相差的日期年月日等 new Date().toString('yyyy-MM-dd'); */
    Date.prototype.toString = function(fmt) {
        const o = {
            "M+": this.getMonth() + 1, //月份           
            "d+": this.getDate(), //日           
            "h+": this.getHours() % 12 == 0 ? 12 : this.getHours() % 12, //小时           
            "H+": this.getHours(), //小时           
            "m+": this.getMinutes(), //分           
            "s+": this.getSeconds(), //秒           
            "q+": Math.floor((this.getMonth() + 3) / 3), //季度           
            "S": this.getMilliseconds() //毫秒           
        };
        const week = {
            "0": "/u65e5",
            "1": "/u4e00",
            "2": "/u4e8c",
            "3": "/u4e09",
            "4": "/u56db",
            "5": "/u4e94",
            "6": "/u516d"
        };
        if (fmt) {} else {
            fmt = 'yyyy/MM/dd HH:mm:ss';
        }
        if (/(y+)/.test(fmt)) {
            fmt = fmt.replace(RegExp.$1, (this.getFullYear() + "").substr(4 - RegExp.$1.length));
        }
        if (/(E+)/.test(fmt)) {
            fmt = fmt.replace(RegExp.$1, ((RegExp.$1.length > 1) ? (RegExp.$1.length > 2 ? "/u661f/u671f" : "/u5468") : "") + week[this.getDay() + ""]);
        }
        for (const k in o) {
            if (new RegExp("(" + k + ")").test(fmt)) {
                fmt = fmt.replace(RegExp.$1, (RegExp.$1.length == 1) ? (o[k]) : (("00" + o[k]).substr(("" + o[k]).length)));
            }
        }
        return fmt;
    };
    //字符串函数扩展
    String.prototype.endWith = String.prototype.endsWith;
    String.prototype.startWith = String.prototype.startsWith;
    String.prototype.eq = (str = '', isOri) => isOri ? hash(this.toString()) == hash(str) : hash(this.toLowerCase()) == hash(str.toLowerCase());
    String.prototype.trim = function(chr) {
        switch (chr) {
            case '/':
            case '\\':
            case '?':
            case '[':
            case ']':
            case '.':
            case '*':
            case '(':
            case ')':
            case '{':
            case '}':
                chr = '\\' + chr;
                break;
        }　　
        return this.replace(isValid(chr) ? new RegExp('(^' + chr + '+)|(' + chr + '+$)') : /(^\s+)|(\s+$)/g, "");　　
    }

}